import 'dart:async';
import 'dart:io';

import 'package:bolt/bolt.dart';
import 'package:bolt/src/utils/utils.dart';
import 'package:meta/meta.dart';

/// {@template bolt_protocol}
/// The underlying protocol for Bolt that can send and receive [DataObject]s.
///
/// Right now it uses UDP directly, this should be doen through provides
/// bindings.
/// {@endtemplate}
abstract class BoltProtocol {
  /// {@macro bolt_protocol}
  BoltProtocol(
    this.address, {
    Logger? logger,
    this.protocolVersion = 1,
  })  : _packetController = StreamController.broadcast(),
        _acknowledgedPacketController = StreamController.broadcast(),
        logger = logger ?? Logger() {
    packets = _packetController.stream;
    acknowledgedPackets = _acknowledgedPacketController.stream;
  }

  /// The registry that holds all the data objects serializers and payload
  /// types.
  final BoltRegistry registry = BoltRegistry();

  /// The address to bind to.
  final Address address;

  /// The logger to use.
  final Logger logger;

  /// The version of the protocol, this is used to determine if the client and
  /// server are compatible.
  final int protocolVersion;

  /// Stream of packets received by the other end.
  late final Stream<Packet> packets;
  final StreamController<Packet> _packetController;

  /// Stream of packets that were acknowledged by the other end.
  late final Stream<Acknowledged<DataObject>> acknowledgedPackets;
  final StreamController<Acknowledged<DataObject>>
      _acknowledgedPacketController;

  RawDatagramSocket? _udpSocket;

  /// The salt to use for a packet sent to [address].
  int retrieveSalt(Address address);

  /// Bind to the [address].
  Future<void> bind() async {
    if (_udpSocket != null) return;

    final hosts = await InternetAddress.lookup(address.host);
    _udpSocket = await RawDatagramSocket.bind(hosts.first, address.port);
    logger.detail('Listening on ${_udpSocket!.address}:${_udpSocket!.port}');

    _udpSocket!.listen((event) {
      if (event == RawSocketEvent.read) {
        final datagram = _udpSocket!.receive();
        if (datagram == null) return;

        final address = Address(datagram.address.address, datagram.port);
        final data = _deserialize(datagram.data, address: address);
        if (data == null) return;
        logger.detail('Received data: $data from $address');
        _packetController.add(Packet(address, data));
      }
    });
  }

  /// Disconnect the binding.
  @mustCallSuper
  void disconnect() {
    _udpSocket?.close();
    _udpSocket = null;
  }

  /// Send a [DataObject] packet to the [address].
  @mustCallSuper
  void rawSend<T extends DataObject, V extends DataResolver<T>>(
    T data,
    Address address, [
    int salt = 0,
  ]) {
    logger.detail('Sending $data to $address');

    _udpSocket?.send(
      _serialize<T, V>(data, saltCheck: salt, address: address),
      // TODO(wolfen): this wont support domain names, only ip addresses
      InternetAddress.tryParse(address.host)!,
      address.port,
    );
  }

  final _packetsSendToAddress = <Address, SequenceBuffer<_SentPacketData>>{};
  final _packetsReceivedFromAddress =
      <Address, SequenceBuffer<_MessageReceivedData>>{};
  final _messagesSentToAddress = <Address, SequenceBuffer<_MessageSentData>>{};
  // final _messagesReceived = SequenceBuffer(1024, _MessageSentData.new);

  /// Serialize the data object into a binary data.
  List<int> _serialize<T extends DataObject, V extends DataResolver<T>>(
    T object, {
    required int saltCheck,
    required Address address,
  }) {
    if (!_packetsSendToAddress.containsKey(address)) {
      _packetsSendToAddress[address] =
          SequenceBuffer(1024, _SentPacketData.new);
    }
    if (!_packetsReceivedFromAddress.containsKey(address)) {
      _packetsReceivedFromAddress[address] =
          SequenceBuffer(1024, _MessageReceivedData.new);
    }
    if (!_messagesSentToAddress.containsKey(address)) {
      _messagesSentToAddress[address] =
          SequenceBuffer(1024, _MessageSentData.new);
    }
    final sentPackets = _packetsSendToAddress[address]!;
    final receivedPackets = _packetsReceivedFromAddress[address]!;
    final sentMessages = _messagesSentToAddress[address]!;

    final serializer = registry.getSerializer<T, V>();
    final data = serializer.serialize(object);

    // The sequence number to use for this packet.
    final packetSequence = sentPackets.sequence;

    // Generate ack bits
    final ack = receivedPackets.sequence - 1;
    var ackBits = 0;
    for (var i = 0; i < 32; i++) {
      final sequence = ack - i;
      if (receivedPackets.find(sequence) != null) {
        ackBits |= 1 << i;
      }
    }

    // Insert ack entry
    final entry = sentPackets.insert(packetSequence);
    if (entry != null) {
      entry.acknowledged = false;
    }

    // Insert message entry
    final messageEntry = sentMessages.insert(packetSequence);
    if (messageEntry != null) {
      messageEntry
        ..timeSent = DateTime.now().millisecondsSinceEpoch
        ..object = object;
    }

    final writer = Payload.write()
      // Header
      ..set(uint16, protocolVersion)
      ..set(uint32, crc32(data))
      ..set(int64, saltCheck)
      ..set(uint16, packetSequence)
      ..set(uint16, ack)
      ..set(uint32, ackBits)

      // Data object
      ..set(uint8, serializer.id)
      ..set(uint16, data.length)
      ..set(Bytes(data.length), data);

    // Inflate the data to 1024 bytes
    return List.filled(1024, 0)..setAll(0, binarize(writer));
  }

  /// Deserialize the data object from a binary data.
  DataObject? _deserialize(
    List<int> data, {
    required Address address,
  }) {
    if (!_packetsSendToAddress.containsKey(address)) {
      _packetsSendToAddress[address] =
          SequenceBuffer(1024, _SentPacketData.new);
    }
    if (!_packetsReceivedFromAddress.containsKey(address)) {
      _packetsReceivedFromAddress[address] =
          SequenceBuffer(1024, _MessageReceivedData.new);
    }
    if (!_messagesSentToAddress.containsKey(address)) {
      _messagesSentToAddress[address] =
          SequenceBuffer(1024, _MessageSentData.new);
    }
    final sentPackets = _packetsSendToAddress[address]!;
    final receivedPackets = _packetsReceivedFromAddress[address]!;
    final sentMessages = _messagesSentToAddress[address]!;

    if (data.length < 1024) {
      throw Exception('Packet is too small');
    }
    final reader = Payload.read(data);

    // Read header
    final protocolVersion = reader.get(uint16);
    final crc = reader.get(uint32);
    final receivedSalt = reader.get(int64);
    final packetSequence = reader.get(uint16);
    final ack = reader.get(uint16);
    var ackBits = reader.get(uint32);

    // Process acknowledgements
    for (var i = 0; i < 32; i++) {
      if (ackBits & 1 == 1) {
        final sequence = ack - i;
        final sentEntry = sentPackets.find(sequence);
        if (sentEntry != null && !sentEntry.acknowledged) {
          final messageEntry = sentMessages.find(sequence);
          if (messageEntry != null) {
            final time = DateTime.now().millisecondsSinceEpoch;
            final latency = time - messageEntry.timeSent;

            _acknowledgedPacketController.add(
              Acknowledged(address, messageEntry.object!, latency),
            );
          }
          sentEntry.acknowledged = true;
        }
      }
      ackBits >>= 1;
    }

    receivedPackets.insert(packetSequence);

    // Read data object
    final id = reader.get(uint8);
    final length = reader.get(uint16);
    final bytes = reader.get(Bytes(length));

    if (protocolVersion != this.protocolVersion) {
      throw Exception('Invalid protocol id: $protocolVersion');
    }

    // Invalid salt, ignore data
    if (receivedSalt != retrieveSalt(address)) return null;
    if (crc != crc32(bytes)) {
      throw Exception('CRC32 checksum failed');
    }

    return registry.getSerializerById(id)?.deserialize(bytes);
  }
}

class _SentPacketData {
  bool acknowledged = false;
}

class _MessageSentData {
  int timeSent = 0;

  DataObject? object;
}

class _MessageReceivedData {}
